/***
 * Cecilia ADL Task Framework
 * Copyright (C) 2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.task.core.control;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.julia.Controller;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.util.Fractal;

public abstract class CompositeExecutionControllerMixin
    implements
      ExecutionController,
      Controller {

  // -------------------------------------------------------------------------
  // Private constructor
  // -------------------------------------------------------------------------

  private CompositeExecutionControllerMixin() {
  }

  // -------------------------------------------------------------------------
  // Fields and methods added and overridden by the mixin class
  // -------------------------------------------------------------------------

  /**
   * Boolean that indicates if the task has already been executed.
   */
  public boolean fcHasBeenExecuted;

  public void initFcController(final InitializationContext ic)
      throws InstantiationException {
    fcHasBeenExecuted = false;
    _super_initFcController(ic);
  }

  public void execute() throws Exception {
    if (fcHasBeenExecuted) return;

    try {
      for (final Component subComponent : scheduleComponents()) {
        System.out.println("Execute component " + subComponent);
        ((ExecutionController) subComponent
            .getFcInterface(ExecutionController.DEFAULT_NAME)).execute();
      }
    } finally {
      fcHasBeenExecuted = true;
    }
  }

  public boolean isFcExecuted() {
    return fcHasBeenExecuted;
  }

  // -------------------------------------------------------------------------
  // internal methods
  // -------------------------------------------------------------------------

  private List<Component> scheduleComponents() {
    final Component[] subComponents = _this_weaveableCC.getFcSubComponents();

    // bindingGraph associates task component to the components that it is bound
    // to
    final Map<Component, Collection<Component>> bindingGraph = new HashMap<Component, Collection<Component>>();

    for (final Component subComponent : subComponents) {
      try {
        subComponent.getFcInterface(ExecutionController.DEFAULT_NAME);
      } catch (final NoSuchInterfaceException e) {
        // subComponent does not have an ExecutionController interface.
        continue;
      }

      final Set<Component> serverComponents = new HashSet<Component>();
      final BindingController subBC;
      try {
        subBC = Fractal.getBindingController(subComponent);
      } catch (final NoSuchInterfaceException e) {
        bindingGraph.put(subComponent, serverComponents);
        continue;
      }

      final String[] itfNames = subBC.listFc();
      for (final String itfName : itfNames) {
        Object serverItf = null;
        try {
          serverItf = subBC.lookupFc(itfName);
        } catch (final NoSuchInterfaceException ignored) {
          // should not append
        }
        if (serverItf == null) continue;

        final Component owner = ((Interface) serverItf).getFcItfOwner();
        if (owner != _this_weaveableC) {
          serverComponents.add(owner);
        }
      }

      bindingGraph.put(subComponent, serverComponents);
    }
    System.out.println("component graph: " + bindingGraph);

    // build the transitive closure of the bindingGraph. i.e. create a graph
    // associating a given component to the set of every component
    // transitively accessible through the bindingGraph.
    final Map<Component, Collection<Component>> transitiveGraph = new HashMap<Component, Collection<Component>>();

    for (final Component subComponent : subComponents) {
      final Set<Component> transitivelyAccessibleComponents = new HashSet<Component>();
      buildTransitiveGraph(subComponent, bindingGraph,
          transitivelyAccessibleComponents, subComponent);
      transitiveGraph.put(subComponent, transitivelyAccessibleComponents);
    }
    System.out.println("transitive component graph: " + transitiveGraph);

    // build a deep first order.
    final List<Component> executeOrder = new ArrayList<Component>(
        subComponents.length);

    for (final Component subComponent : subComponents) {
      buildDeepFirstOrder(transitiveGraph, executeOrder, subComponent);
    }

    return executeOrder;
  }

  private void buildTransitiveGraph(final Component initialComponent,
      final Map<Component, Collection<Component>> bindingGraph,
      final Set<Component> transitivelyAccessibleComponents,
      final Component component) {
    final Collection<Component> serverComponents = bindingGraph.get(component);
    if (serverComponents == null) return;
    for (final Component serverComponent : serverComponents) {
      if (serverComponent == initialComponent) {
        // TODO Cycle !! throw an exception.
      }

      if (transitivelyAccessibleComponents.add(serverComponent)) {
        // serverComponent added in the set, call recursively
        buildTransitiveGraph(initialComponent, bindingGraph,
            transitivelyAccessibleComponents, serverComponent);
      }
    }
  }

  private void buildDeepFirstOrder(
      final Map<Component, Collection<Component>> graph,
      final List<Component> executeOrder, final Component component) {
    final Collection<Component> dependencies = graph.remove(component);
    if (dependencies != null) {
      for (final Component dependency : dependencies) {
        buildDeepFirstOrder(graph, executeOrder, dependency);
      }
      System.out.println("Add in list " + component);
      executeOrder.add(component);
    }
  }

  // -------------------------------------------------------------------------
  // Fields and methods required by the mixin class in the base class
  // -------------------------------------------------------------------------

  /**
   * The {@link Component} interface of the component to which this controller
   * object belongs.
   */
  public Component         _this_weaveableC;

  /**
   * The {@link ContentController} interface of the component to which this
   * controller object belongs.
   */
  public ContentController _this_weaveableCC;

  /**
   * @see Controller#initFcController
   */
  public abstract void _super_initFcController(InitializationContext ic)
      throws InstantiationException;

}
